from os import path
import sys
from importlib import reload


def add_path_(p, pos=0):
    p = path.abspath(path.expanduser(p))
    if p in sys.path:
        return
    if pos is None:
        sys.path.append(p)
    else:
        sys.path.insert(pos, p)


def add_path(*ps, pos=0):
    for p in ps:
        add_path_(p, pos=pos)
    return len(sys.path)



import numpy as np
from bgbb import BGBB
from bgbb.sql.bgbb_udfs import (
    add_p_th,
    mk_n_returns_udf,
    mk_p_alive_udf,
    mk_p_th_exp_udf,
)
from bgbb.wrappers import to_abgd_od
from lifetimes.datasets import load_donations
from lifetimes.fitters.beta_geo_beta_binom_fitter import BetaGeoBetaBinomFitter
from pandas import DataFrame
from pytest import fixture

df = fixture(load_donations)


@fixture
def dfs(df, spark):
    return spark.createDataFrame(df)


@fixture
def params():
    return [1.20, 0.75, 0.66, 2.78]


@fixture
def bgbb(params):
    return BGBB(params=params)


@fixture
def ref_model(params, df):
    reference_model = BetaGeoBetaBinomFitter()
    reference_model.data = df
    reference_model.params_ = to_abgd_od([1.20, 0.75, 0.66, 2.78])
    return reference_model


@fixture
def p_alive(bgbb, params):
    return mk_p_alive_udf(bgbb, params=params, alive_n_days_later=0)


@fixture
def n_returns(bgbb, params):
    return mk_n_returns_udf(bgbb, params=params, return_in_next_n_days=14)


@fixture
def p_th_est_udfs(bgbb, params):
    return mk_p_th_exp_udf(bgbb, params=params)


def test_p_alive(dfs, p_alive, ref_model):
    pa_ref = ref_model.conditional_probability_alive(0)

    df2 = dfs.withColumn(
        "P_alive", p_alive(dfs.frequency, dfs.recency, dfs.n)
    ).toPandas()

    assert np.allclose(df2.P_alive, pa_ref)


def test_n_returns(dfs, n_returns, ref_model):
    p14_ref = ref_model.conditional_expected_number_of_purchases_up_to_time(14)

    df2 = dfs.withColumn(
        "P14", n_returns(dfs.frequency, dfs.recency, dfs.n)
    ).toPandas()

    assert np.allclose(df2.P14, p14_ref)


def test_p_th_est_udf(dfs, p_th_est_udfs):
    """Just test structure and support of this function, since lifetimes
    doesn't have this function.
    """
    p_th_est, _, _ = p_th_est_udfs
    df = dfs.withColumn(
        "p_th", p_th_est(dfs.frequency, dfs.recency, dfs.n)
    ).toPandas()
    p_th_srs = df["p_th"]
    p, th = zip(*p_th_srs)
    p_th_df = DataFrame(dict(p=p, th=th))
    assert p_th_df.eval(
        "0 <= p <= 1 & 0 <= th <= 1"
    ).all(), "This function returns probabilities"


def test_add_p_th(dfs, bgbb, params):
    """
    Check that, given same # opportunities to return,
    higher frequency yields higher estimates of p.
    Also check that columns are as expected, and estimated
    probabilities are between [0, 1].
    """
    p_th_df = add_p_th(
        bgbb, dfs=dfs, params=params, fcol="frequency", rcol="Recency", ncol="N"
    ).toPandas()
    max_opp = p_th_df.n.min()  # noqa: F841
    mean_p_freq = (
        p_th_df.query("n == @max_opp & frequency > 0")
        .groupby("frequency")
        .p.mean()
    )
    # Looks something like:
    # frequency
    # 1    0.383420
    # 2    0.510549
    # 3    0.619845
    # 4    ...
    assert (
        mean_p_freq.is_monotonic_increasing
    ), "As frequency increases, p should as well"
    assert set(p_th_df) == {"frequency", "recency", "n", "n_custs", "p", "th"}
    assert p_th_df.eval(
        "0 <= p <= 1 & 0 <= th <= 1"
    ).all(), "This function returns probabilities"

    none_params = add_p_th(
        bgbb, dfs=dfs, params=None, fcol="frequency", rcol="Recency", ncol="N"
    ).toPandas()
    return none_params



"""
Most of this stuff from
https://github.com/mozilla/python_mozetl/blob/master/tests/conftest.py
"""
import pyspark
import pytest
from pytest import fixture


@fixture
def spark():
    spark = pyspark.sql.SparkSession.builder.appName("test").getOrCreate()
    yield spark
    spark.stop()


class DataFrameFactory:
    """Create a dataframe given a base dictionary and schema."""

    def __init__(self, spark_session):
        self.spark = spark_session

    def create_dataframe(self, snippets, base, schema=None):
        """Generate a dataframe in the shape of the base dictionary where every
        row has column values overwritten by the snippets.

        :snippets list[dict]: a list of fields to overwrite in the base
            dictionary
        :base dict: a base instantiation of a row in the dataset
        :schema pyspark.sql.types.StructType: schema for the dataset
        """
        # the dataframe should have at least one item
        if not snippets:
            snippets = [dict()]

        samples = []
        for snippet in snippets:
            sample = base.copy()
            if snippet is None:
                snippet = {}
            sample.update(snippet)
            samples.append(sample)

        # if no schema is provided, the schema will be inferred
        return self.spark.createDataFrame(samples, schema)

    def create_dataframe_with_key(
        self, snippets, base, key, key_func=None, schema=None
    ):
        """Generate dataframe with autoincrementing key function"""

        def generate_keys():
            num = 0
            while True:
                yield str(num)
                num += 1

        # default key function
        if not key_func:
            key_func = generate_keys

        if not snippets:
            snippets = [dict()]

        # update each snippet with new key
        gen = key_func()
        for i in range(len(snippets)):
            snippets[i].update({key: next(gen)})

        return self.create_dataframe(snippets, base, schema)


@pytest.fixture()
def dataframe_factory(spark):
    """A factory object for generating test datasets.

    This fixture provides methods for generating data from a base document
    and a list of columns to modify.

    Example usage:

        from functools import partial
        import pytest

        @pytest.fixture
        def base_document():
            return {"uid": 1}

        @pytest.fixture
        def generate_data(dataframe_factory, base_document):
            return partial(dataframe_factory.create_dataframe,
                base=base_document)

        def test_3_unique_uids(generate_data):
            data = generate_data([
                {"uid": 1},
                {"uid": 2},
                {"uid": 3}
            ])
            assert data.count() == 3
    """
    return DataFrameFactory(spark)



import datetime as dt

import pandas as pd
from pytest import raises, fixture

from bgbb.sql.sql_utils import to_sql_list, mk_time_params, to_samp_ids

mod_win1 = 90
mod_win2 = 120

ho_win1 = 14
ho_win2 = 21


@fixture
def r1():
    return mk_time_params(
        ho_win=ho_win1, model_win=mod_win1, ho_start="2018-08-01"
    )


@fixture
def r2():
    return mk_time_params(
        ho_win=ho_win2, model_win=mod_win2, ho_start="2001-11-01"
    )


def check_model_range(
    model_start_date=None, ho_start_date=None, model_win=None
):
    """
    if model input period is 14 days, make sure all 14 days are present
    in the range; Basically, ensure inclusive range.
    """
    mod_range = pd.date_range(
        start=model_start_date, periods=model_win, freq="D"
    )
    assert len(mod_range) == model_win
    assert mod_range[0].to_pydatetime().date() == model_start_date
    assert (
        mod_range[-1].to_pydatetime().date() + dt.timedelta(days=1)
    ) == ho_start_date


def check_ho_range(ho_start_date, ho_last_date, ho_win):
    ho_range = pd.date_range(start=ho_start_date, periods=ho_win, freq="D")
    assert len(ho_range) == ho_win
    assert ho_range[0].to_pydatetime().date() == ho_start_date
    assert ho_range[-1].to_pydatetime().date() == ho_last_date


def test_to_sql_list():
    assert to_sql_list(["GB"]) == "'GB'"
    assert to_sql_list(["GB", "US", "IN"]) == "'GB', 'US', 'IN'"
    assert to_sql_list([1, 2, 3]) == "1, 2, 3"


def test_mk_time_param_model_range(r1, r2):
    check_model_range(
        model_start_date=r1.model_start_date,
        ho_start_date=r1.ho_start_date,
        model_win=mod_win1,
    )

    check_model_range(
        model_start_date=r2.model_start_date,
        ho_start_date=r2.ho_start_date,
        model_win=mod_win2,
    )

    # Check this test is actually doing something with wrong
    # `model_win` arg
    raises(
        AssertionError,
        check_model_range,
        model_start_date=r2.model_start_date,
        ho_start_date=r2.ho_start_date,
        model_win=mod_win1,
    )


def test_ho_range(r1, r2):
    check_ho_range(r1.ho_start_date, r1.ho_last_date, ho_win1)
    check_ho_range(r2.ho_start_date, r2.ho_last_date, ho_win2)
    raises(
        AssertionError,
        check_ho_range,
        r2.ho_start_date,
        r2.ho_last_date,
        ho_win1,
    )


def test_to_samp_ids():
    assert to_samp_ids([42]) == "'42'"
    assert to_samp_ids([]) == ''
    assert to_samp_ids(["42"]) == "'42'"
    assert to_samp_ids(["42", 1]) == "'42', '1'"



import numpy as np
from bgbb.bgbb_likelihood import nb_loglikelihood


def test_nb_loglikelihood():
    res = nb_loglikelihood([0.1, 0.1, 0.2, 0.3], np.r_[3], np.r_[4], np.r_[5])
    assert np.allclose(res, [-6.31032992])



import numpy as np
from numba import njit, prange
from numpy import log, exp, logaddexp
from scipy.special import betaln

from bgbb.numba_special import nb_lbeta_vec12, nb_lbeta_vec2


def get_vals(x):
    return x.values if hasattr(x, "values") else x


@njit
def nbsum_scalar(
    x, tx, recency_T, alpha, beta, gamma, delta, J, beta_ab, beta_gd
):
    if recency_T <= -1:
        return -np.inf

    j = J[: recency_T + 1]
    return log(
        np.sum(
            exp(
                nb_lbeta_vec2(alpha + x, beta + tx - x + j)
                - beta_ab
                + nb_lbeta_vec2(gamma + 1, delta + tx + j)
                - beta_gd
            )
        )
    )


@njit
def nbsum_(x, tx, recency_T, a, b, g, d, J, ab, gd):
    N = len(x)
    res = np.empty(N)
    for i in range(N):
        res[i] = nbsum_scalar(x[i], tx[i], recency_T[i], a, b, g, d, J, ab, gd)
    return res


@njit(parallel=True)
def nbsum_p(x, tx, recency_T, a, b, g, d, J, ab, gd):
    N = len(x)
    res = np.empty(N)
    for i in prange(N):
        res[i] = nbsum_scalar(x[i], tx[i], recency_T[i], a, b, g, d, J, ab, gd)
    return res


def nbsum(x, tx, recency_T, a, b, g, d, J, ab, gd, parallel=True):
    f = nbsum_p if parallel else nbsum_
    return f(x, tx, recency_T, a, b, g, d, J, ab, gd)


def _nb_loglikelihood(alpha, beta, gamma, delta, x, tx, T, parallel=True):
    """Loglikelihood for optimizer."""
    # Constrain the params to be > 0
    if min([alpha, beta, gamma, delta]) <= 0:
        return np.full_like(x, -np.inf)

    beta_ab = betaln(alpha, beta)
    beta_gd = betaln(gamma, delta)

    recency_T = T - tx - 1
    xa, txa, Ta, recency_Ta = map(get_vals, [x, tx, T, recency_T])

    indiv_loglike = (
        nb_lbeta_vec12(alpha + xa, beta + Ta - xa)
        - beta_ab
        + nb_lbeta_vec2(gamma, delta + Ta)
        - beta_gd
    )

    J = np.arange(recency_T.max() + 1)

    s = nbsum(
        xa,
        txa,
        recency_Ta,
        alpha,
        beta,
        gamma,
        delta,
        J,
        beta_ab,
        beta_gd,
        parallel=parallel,
    )

    indiv_loglike = logaddexp(indiv_loglike, s)

    return indiv_loglike


############
# Wrappers #
############
def nb_loglikelihood(params, x, tx, T, parallel=True):
    alpha, beta, gamma, delta = params
    return _nb_loglikelihood(
        alpha, beta, gamma, delta, x, tx, T, parallel=parallel
    )


def nb_loglikelihood_df(
    params, df, rfn_names=False, parallel=True, nll=False, ncusts=None
):
    columns = ["frequency", "recency", "n"] if rfn_names else ["x", "tx", "T"]
    x, tx, T = [df[c] for c in columns]
    alpha, beta, gamma, delta = params
    ll = _nb_loglikelihood(
        alpha, beta, gamma, delta, x, tx, T, parallel=parallel
    )
    if nll:
        if ncusts is None:
            print("WARNING: ncusts not passed")
            ncusts = 1
        return -np.mean(ll * ncusts)
    return ll



from typing import Optional

from bgbb import BGBB
from bgbb.bgbb_utils import AbgdParams
from pandas import Series
from pyspark.sql.functions import PandasUDFType, pandas_udf, col as C


def mk_p_alive_udf(bgbb: BGBB, params=None, alive_n_days_later=0):
    @pandas_udf("double", PandasUDFType.SCALAR)
    def p_alive(frequency, recency, n, params=params):
        arr = bgbb.cond_prob_alive_nb(
            frequency,
            recency,
            n,
            params=params,
            n_days_later=alive_n_days_later,
        )
        return Series(arr)

    return p_alive


def mk_n_returns_udf(bgbb: BGBB, params=None, return_in_next_n_days=14):
    @pandas_udf("double", PandasUDFType.SCALAR)
    def n_returns(f, r, t):
        arr = bgbb.cond_exp_rets_till_nb(return_in_next_n_days, f, r, t, params)
        return Series(arr)

    return n_returns


def mk_p_th_exp_udf(bgbb: BGBB, params: Optional[AbgdParams] = None):
    "Compute expected value of latent params p, theta"

    params = params or bgbb.Params

    @pandas_udf("array<double>")
    def p_th_est(f, r, t):
        p_est, th_est = bgbb._expec_p_th(*params, f, r, t)
        return Series([[p, th] for p, th in zip(p_est, th_est)])

    @pandas_udf("double", PandasUDFType.SCALAR)
    def p_est(p_th):
        return Series([x[0] for x in p_th])

    @pandas_udf("double", PandasUDFType.SCALAR)
    def th_est(p_th):
        return Series([x[1] for x in p_th])

    return p_th_est, p_est, th_est


def add_p_th(
    bgbb: BGBB, dfs, params=None, fcol="Frequency", rcol="Recency", ncol="N"
):
    """
    Use udfs from `mk_p_th_exp_udf` to create columns `p` and `th`
    from intermediate computed column.
    """
    p_th_est, p_est, th_est = mk_p_th_exp_udf(bgbb, params=params)
    res = (
        dfs.withColumn("p_th", p_th_est(C(fcol), C(rcol), C(ncol)))
        .withColumn("p", p_est(C("p_th")))
        .withColumn("th", th_est(C("p_th")))
        .drop("p_th")
    )
    return res



from collections import OrderedDict, namedtuple
from itertools import count
from typing import List, Optional, Sequence, Tuple

import numpy as np
import numpy.random as nr
from numba import njit
from pandas import DataFrame

Prob = float


###############
# Simulations #
###############
def gen_probs(abgd: List[float], n=10):
    pa, pb, ta, tb = abgd
    p = nr.beta(pa, pb, size=n)
    th = nr.beta(ta, tb, size=n)
    return p, th


def gen_buy_die(
    n_opps,
    n_users,
    abgd: List[float],
    p_th: Optional[Tuple[Sequence[Prob], Sequence[Prob]]] = None,
    n_opps_name="n_opps",
    seed=0,
):
    """
    Given `n_opps` window size of opportunity, simulate Buy 'Til You Die process
    for `n_users` users.
    If arrays for latent variables p_th = (p, theta) are not passed, then these
    are drawn from beta distribution.
    """
    nr.seed(seed)
    if p_th is None:
        p, th = gen_probs(abgd, n=n_users)
    else:
        p, th = p_th
    txs, xs = np.empty_like(p, dtype=np.int), np.empty_like(p, dtype=np.int)
    bern = lambda p: nr.binomial(1, p, size=n_opps)

    for i, pi, thi in zip(count(), p, th):
        buys, dies = bern(pi), bern(thi)
        xs[i], txs[i] = get_x_tx(buys, dies)

    ret = DataFrame(
        OrderedDict([("p", p), ("th", th), ("frequency", xs), ("recency", txs)])
    ).assign(**{n_opps_name: n_opps})
    return ret


@njit
def get_x_tx(buys, dies) -> Tuple[int, int]:
    """
    Converts simulated 'buy'/'die' events into frequency/recency
    counts.
    """
    x = 0
    tx = 0
    for opp, buy, die in zip(range(1, len(buys) + 1), buys, dies):
        if die:
            break
        if buy:
            x += 1
            tx = opp
    return x, tx


_AbgdParams = namedtuple("AbgdParams", ["a", "b", "g", "d"])


class AbgdParams(_AbgdParams):
    def mod_param(self, **par_fns):
        dct = self._asdict().copy()
        for par_letter, f in par_fns.items():
            dct[par_letter] = f(dct[par_letter])
        return self.from_dct(dct)

    @property
    def greek_dct(self):
        dct = self._asdict().copy()
        for name, letter in zip(["α", "β", "γ", "δ"], "abgd"):
            dct[name] = dct.pop(letter)
        return dct

    @classmethod
    def from_dct(cls, dct):
        return cls(*[dct[k] for k in "abgd"])

    def __repr__(self):
        st_repr = ", ".join(
            "{}: {:.1f}".format(k, v) for k, v in self.greek_dct.items()
        )
        return "BGBB Hyperparams <{}>".format(st_repr)



import math

import numpy as np
from numba import njit
from numba import vectorize
from numpy import exp


@njit
def nb_logsumexp(xs):
    mx = np.max(xs)
    s = 0
    for x in xs:
        s += np.exp(x - mx)
    return mx + np.log(s)


# TODO: delete?
# @njit
@vectorize
def lgamma(x):
    return math.lgamma(x)


@njit
def nb_lbeta(z, w):
    """See: https://www.mathworks.com/help/matlab/ref/betaln.html"""
    return lgamma(z) + lgamma(w) - lgamma(z + w)


@njit
def nb_lbeta_vec2(z, w: np.array):
    N = len(w)
    res = np.empty(N)
    for i in range(N):
        res[i] = nb_lbeta(z, w[i])
    return res


@njit
def nb_lbeta_vec12(z: np.array, w: np.array):
    N = len(w)
    res = np.empty(N)
    for i in range(N):
        res[i] = nb_lbeta(z[i], w[i])
    return res


##############
# Components #
##############
@njit
def p_alive_exp_p1_p2(params, xa, na, n_days_later, p3):
    """For cond_prob_alive_nb"""
    alpha, beta, gamma, delta = params
    p1nb = nb_lbeta(alpha + xa, beta + na - xa) - nb_lbeta(alpha, beta)
    p2nb = nb_lbeta(gamma, delta + na + n_days_later) - nb_lbeta(gamma, delta)
    return exp(p1nb + p2nb) / exp(p3)


@njit
def cond_exp_rets_till_p2345(t, frequency, recency, n, params, p1):
    alpha, beta, gamma, delta = params
    x = frequency

    p2 = exp(nb_lbeta(alpha + x + 1, beta + n - x) - nb_lbeta(alpha, beta))
    p3 = delta / (gamma - 1) * exp(lgamma(gamma + delta) - lgamma(1 + delta))
    p4 = exp(lgamma(1 + delta + n) - lgamma(gamma + delta + n))
    p5 = exp(lgamma(1 + delta + n + t) - lgamma(gamma + delta + n + t))

    return p1 * p2 * p3 * (p4 - p5)



from collections import OrderedDict
from functools import wraps
import inspect
from typing import Dict, Union, List
import numpy as np

from lifetimes import BetaGeoBetaBinomFitter

abgd_names = "alpha beta gamma delta".split()


def unload(dct: Dict, ks: Union[List[str], str]) -> List[float]:
    if isinstance(ks, str):
        ks = ks.split()
    return [dct[k] for k in ks]


def frt(f):
    """Wraps lifetimes model methods that potentially
    take 'frequency', 'recency', 'T' as params,
    replace the func w/ one that takes a dataframe
    with these arguments as column names.
    """
    frt_params = {"frequency", "recency", "T", "n", "n_custs"}
    sig = inspect.signature(f)
    params = list(sig.parameters)

    @wraps(f)
    def wrapper(self, data, **k):
        k.update({p: data[p] for p in params if p in frt_params})
        return f(self, **k)

    return wrapper


def to_abgd_od(params):
    # type: (Union["OrderedDict[str, float]", List[float]]) -> OrderedDict[str, float]
    if isinstance(params, OrderedDict):
        return params
    return OrderedDict(zip(abgd_names, params))


def module_parameter_list(params, mod):
    return params or list(mod.params_.values())


class Rfn:
    """TODO: document
    """

    def __init__(self, mod):
        self.mod = mod

    def cond_prob_alive(
        self, df, params: List[float] = None, n_days_later=0, nb=True
    ):
        params = module_parameter_list(params, self.mod)
        frequency, recency, n = unload(df, "frequency recency n")
        kw = dict(
            frequency=frequency,
            recency=recency,
            n=n,
            params=params,
            n_days_later=n_days_later,
        )
        if nb:
            return self.mod.cond_prob_alive_nb(**kw)
        return self.mod.cond_prob_alive(**kw)

    @wraps(BetaGeoBetaBinomFitter.fit)
    def fit(self, df, **kw):
        frequency, recency, n = unload(df, "frequency recency n")
        if "n_custs" in df:
            n_custs = df["n_custs"]
        else:
            print("Warning: no n_custs column")
            n_custs = np.ones_like(frequency)
        return self.mod.fit(frequency, recency, n, n_custs, **kw)

    @wraps(BetaGeoBetaBinomFitter._loglikelihood)
    def _loglikelihood(self, df, params=None, parallel=True):
        x, tx, T = unload(df, "frequency recency n")
        params = module_parameter_list(params, self.mod)
        return self.mod._loglikelihood(params, x, tx, T, parallel=parallel)

    @wraps(
        BetaGeoBetaBinomFitter.conditional_expected_number_of_purchases_up_to_time
    )
    def cond_exp_rets_till(self, df, n_days_later, params=None, nb=False):
        x, tx, n = unload(df, "frequency recency n")
        params = module_parameter_list(params, self.mod)
        kw = dict(t=n_days_later, frequency=x, recency=tx, n=n, params=params)
        if nb:
            return self.mod.cond_exp_rets_till_nb(**kw)
        return self.mod.cond_exp_rets_till(**kw)



import re
from setuptools import find_packages, setup


def read_version():
    "https://stackoverflow.com/a/7071358/386279"
    VERSIONFILE = "bgbb/_version.py"

    with open(VERSIONFILE, "rt") as fp:
        verstrline = fp.read()
    VSRE = r"^__version__ = ['\"]([^'\"]*)['\"]"
    mo = re.search(VSRE, verstrline, re.M)
    if mo:
        verstr = mo.group(1)
    else:
        raise RuntimeError(
            "Unable to find version string in %s." % (VERSIONFILE,)
        )
    return verstr


verstr = read_version()

setup(
    name="bgbb",
    version=verstr,
    packages=find_packages(),
    install_requires=[
        "lifetimes==0.9",
        # See Readme for databricks library restrictions
        "numba>=0.34,<=0.38",
        "scipy",
        "pandas",
        "numpy",
        "pyarrow",
        "pyspark",
    ],
    license=["Apache 2.0", "MIT"],
    extras_require={"test": ["pytest"]},
    classifiers=["Programming Language :: Python :: 3.6"],
)



from typing import List, Optional, Sequence, Tuple, Union

import numpy as np
import pandas as pd
from lifetimes import BetaGeoBetaBinomFitter
from numpy import exp
from scipy.special import betaln, gammaln

from bgbb.bgbb_likelihood import nb_loglikelihood
from bgbb.bgbb_utils import AbgdParams
from bgbb.numba_special import cond_exp_rets_till_p2345, p_alive_exp_p1_p2
from bgbb.wrappers import Rfn, frt, to_abgd_od

Prob = float  # float: [0, 1]


def as_array(s: Union[np.array, pd.Series]):
    try:
        return s.values
    except AttributeError:
        return s


class BGBB(BetaGeoBetaBinomFitter):
    def __init__(
        self, penalizer_coef=0, params: Optional[List[float]] = None, data=None
    ):
        if params is not None:
            self.params_ = to_abgd_od(params)
            self.Params = AbgdParams(*self.params_.values())
        if data is not None:
            self.data = data
        self.rfn = Rfn(self)
        super().__init__(penalizer_coef=penalizer_coef)

    # conditional expected number of events up to time
    c_exp_num_ev_to = frt(
        BetaGeoBetaBinomFitter.conditional_expected_number_of_purchases_up_to_time
    )

    _loglikelihood = staticmethod(nb_loglikelihood)
    old_loglikelihood = BetaGeoBetaBinomFitter._loglikelihood

    @classmethod
    def _negative_log_likelihood(
        cls, params, frequency, recency, n, n_custs, penalizer_coef=0
    ):
        penalizer_term = penalizer_coef * sum(np.asarray(params) ** 2)
        return (
            -np.mean(
                cls._loglikelihood(params, frequency, recency, n) * n_custs
            )
            + penalizer_term
        )

    @classmethod
    def cond_prob_alive(self, frequency, recency, n, params, n_days_later=0):
        """
        Version to be able to explicitly pass RFN columns.
        Conditional probability alive.

        From https://github.com/CamDavidsonPilon/lifetimes:
        Conditional probability customer is alive at transaction opportunity
        n + n_days_later.

        .. math:: P(alive at n + n_days_later
                    |alpha, beta, gamma, delta, frequency, recency, n)

        See (A10) in Fader and Hardie 2010.

        Parameters
        ----------
        n_days_later: array_like
            transaction opportunities

        Returns
        -------
        array_like
            alive probabilities

        """
        alpha, beta, gamma, delta = params
        x, tx = frequency, recency

        p1 = betaln(alpha + x, beta + n - x) - betaln(alpha, beta)
        p2 = betaln(gamma, delta + n + n_days_later) - betaln(gamma, delta)
        p3 = self._loglikelihood(params, x, tx, n)

        return exp(p1 + p2) / exp(p3)

    @classmethod
    def cond_prob_alive_nb(cls, frequency, recency, n, params, n_days_later=0):
        "Numba Version"
        x, tx = frequency, recency
        xa, _, na = map(as_array, [x, tx, n])
        p3 = cls._loglikelihood(params, x, tx, n)

        return p_alive_exp_p1_p2(params, xa, na, n_days_later, p3)

    def cond_exp_rets_till(self, t, frequency, recency, n, params):
        x, tx = frequency, recency
        alpha, beta, gamma, delta = params

        p1 = 1 / exp(self._loglikelihood(params, x, tx, n))
        p2 = exp(betaln(alpha + x + 1, beta + n - x) - betaln(alpha, beta))
        p3 = (
            delta
            / (gamma - 1)
            * exp(gammaln(gamma + delta) - gammaln(1 + delta))
        )
        p4 = exp(gammaln(1 + delta + n) - gammaln(gamma + delta + n))
        p5 = exp(gammaln(1 + delta + n + t) - gammaln(gamma + delta + n + t))

        return p1 * p2 * p3 * (p4 - p5)

    def cond_exp_rets_till_nb(self, t, frequency, recency, n, params):
        xa, txa, na = map(as_array, [frequency, recency, n])

        p1 = 1 / exp(self._loglikelihood(params, xa, txa, na))
        ret = cond_exp_rets_till_p2345(t, xa, txa, na, params, p1)
        return ret

    @classmethod
    def _log_expec_p_th(cls, a: float, b: float, g: float, d: float, f, r, n):
        l_ap1 = cls._loglikelihood([a + 1, b, g, d], f, r, n)
        l_gp1 = cls._loglikelihood([a, b, g + 1, d], f, r, n)
        ll = cls._loglikelihood([a, b, g, d], f, r, n)

        l_a_frac = np.log(a) - np.log(a + b)
        l_g_frac = np.log(g) - np.log(g + d)

        return (l_a_frac + l_ap1 - ll), (l_g_frac + l_gp1 - ll)

    @classmethod
    def _expec_p_th(cls, a, b, g, d, f, r, n):
        lp_exp, lth_exp = cls._log_expec_p_th(a, b, g, d, f, r, n)
        return np.exp(lp_exp), np.exp(lth_exp)

    def latent_variable_mean(
        self, f, r, n
    ) -> Tuple[Sequence[Prob], Sequence[Prob]]:
        """
        Expected values of users' latent transaction and dropout probabilities
        eqns (13), (14) from "Customer-Base Analysis in a Discrete-Time Noncontractual Setting,"
        Fader 2009.
        """
        return self._expec_p_th(*self.Params, f, r, n)



import datetime as dt
from typing import List, Union

import pandas as pd
from pandas.compat import lmap


S3_DAY_FMT = "%Y%m%d"
S3_DAY_FMT_DASH = "%Y-%m-%d"


def to_s3_fmt(date):
    return date.strftime(S3_DAY_FMT)


def to_samp_ids(samp_ids: Union[List[int], int]) -> str:
    """
    iter of ints to SQL string version for main_summary
    >>> to_samp_ids([0, 1, 2])
    "'0', '1', '2'"
    """
    if not isinstance(samp_ids, (list, range)):
        samp_ids = [samp_ids]
    samp_ids_i = lmap(int, samp_ids)  # type: List[int]
    invalid_sample = set(samp_ids_i) - set(range(100))
    if invalid_sample:
        raise ValueError(
            "{} is outside of the valid range [0, 99]".format(invalid_sample)
        )
    return to_sql_list(map(str, samp_ids_i))


def to_sql_list(xs):
    """stringify lists for SQL queries
    >>> to_sql_list([1, 2, 3]) == '1, 2, 3'
    """

    def to_sql_literal(x):
        if isinstance(x, str):
            return "'{}'".format(x)
        return str(x)

    res = ", ".join(map(to_sql_literal, xs))
    return res


def insert_country(
    q, insert_before="{sample_comment}", countries: List[str] = ["GB"]
):
    "Insert country restriction into SQL string for testing"
    i = q.find(insert_before)
    to_insert = "AND country IN ({})\n      ".format(to_sql_list(countries))
    return q[:i] + to_insert + q[i:]


def mk_time_params(ho_win=14, model_win=90, ho_start="2018-08-01"):
    """
    Return container whose attributes are holdout and model input
    date ranges, specified by a training window `model_win`,
    holdout evaluation window `ho_win` and holdout start date `ho_start`
    (day after last day in model window).
    """

    def r():
        pass

    r.ho_start_datet = pd.to_datetime(ho_start)
    r.ho_start_date = r.ho_start_datet.date()
    r.ho_last_date = r.ho_start_date + dt.timedelta(days=ho_win - 1)
    r.model_start_date = r.ho_start_date - dt.timedelta(days=model_win)

    # Str format
    mod_ho_ev = [r.model_start_date, r.ho_start_date, r.ho_last_date]
    # mod_ho_ev_str
    (r.model_start_date_str, r.ho_start_date_str, r.ho_last_date_str) = map(
        to_s3_fmt, mod_ho_ev
    )

    # r.__dict__.update(locals())
    return r


base_query = """
with cid_day as (
    SELECT
        C.client_id
        , C.sample_id
        , C.submission_date_s3
        , from_unixtime(unix_timestamp(C.submission_date_s3, 'yyyyMMdd'),
                        'yyyy-MM-dd') AS sub_date
    FROM clients_daily C
    WHERE
        app_name = 'Firefox'
        AND channel = 'release'
      {sample_comment}  AND sample_id in ({sample_ids})
)

-- clients_daily aggregates from window *before*
-- the holdout date
, cid_model as (
    SELECT
        C.client_id
        , C.sample_id
        , MIN(C.sub_date) AS Min_day
        , MAX(C.sub_date) AS Max_day
        , COUNT(*) AS X
    FROM cid_day C
    WHERE
      C.submission_date_s3 >= '{model_start_date_str}'
      AND C.submission_date_s3 < '{ho_start_date_str}'
    GROUP BY 1, 2
)

, cid_holdout as (
    SELECT
        C.client_id
        , COUNT(*) AS N_holdout
    FROM cid_day C
    WHERE
      C.submission_date_s3 >= '{ho_start_date_str}'
      AND C.submission_date_s3 <= '{ho_last_date_str}'
    GROUP BY 1
)

, rec_freq as (
    SELECT
        C.client_id
        , C.sample_id
        , datediff(C.Max_day, Min_day) AS Recency
        , X - 1 AS Frequency
        -- N: # opportunities to return
        , datediff('{ho_start_date}', Min_day) - 1  AS N
        , C.Max_day
        , C.Min_day
    FROM cid_model C
)

, rec_freq_holdout as (
  SELECT R.*
        , coalesce(H.N_holdout, 0) AS N_holdout
  FROM rec_freq R
  LEFT JOIN cid_holdout H
    ON R.client_id = H.client_id
)

SELECT * FROM {qname}
"""


# TODO: test both holdout=True and False
def mk_rec_freq_q(
    q=None,
    holdout=False,
    model_start_date_str: str = None,
    pcd=None,
    sample_ids: List[int] = [1],
    **k
):
    """
    holdout: pull # of returns in holdout period?
    """
    qname = "rec_freq_holdout" if holdout else "rec_freq"
    sample_ids_str = to_samp_ids(sample_ids)  # type: str

    kw = dict(
        model_start_date_str=model_start_date_str,
        sample_ids=sample_ids_str,
        sample_comment="" if sample_ids else "--",
        qname=qname,
        pcd=pcd,
    )
    kw.update(k)
    kw = {k: v for k, v in kw.items() if v is not None}
    return q.format(**kw)


def run_rec_freq_spk(
    spark,
    rfn_base_query=base_query,
    ho_win=14,
    model_win=90,
    holdout=False,
    sample_ids: List[int] = [],
    ho_start="2018-08-01",
    ho_days_in_future=None,
):
    """
    holdout: whether to pull # of returns in holdout period. Useful
        for evaluating model.
    ho_days_in_future: int?; if 1, set the first day in holdout period
        to be tomorrow. Negative numbers will set it to before today.
    """
    if ho_days_in_future is not None:
        ho_start = dt.date.today() + dt.timedelta(days=ho_days_in_future)
    r = mk_time_params(ho_win=ho_win, model_win=model_win, ho_start=ho_start)
    r.q = mk_rec_freq_q(
        q=rfn_base_query,
        holdout=holdout,
        # ignore_pcd=ignore_pcd,
        sample_ids=sample_ids,
        model_start_date_str=r.model_start_date_str,
        pcd=r.model_start_date,
        ho_start_date=r.ho_start_date,
        ho_last_date_str=r.ho_last_date_str,
        ho_start_date_str=r.ho_start_date_str,
    )
    #     print(r.q)
    dfs = spark.sql(r.q)
    return dfs, r.q


def reduce_rec_freq_spk(dfs, rfn_cols=["Recency", "Frequency", "N"]):
    """Reduce r/f/n spark dataframe to r/f/n pattern count.
    This can be used for fitting.
    """
    return dfs.groupby(rfn_cols).count().withColumnRenamed("count", "n_users")


def rec_freq_spk2pandas(df_spk):
    df = df_spk.toPandas()
    df = df.assign(
        Max_day=lambda x: pd.to_datetime(x.Max_day),
        Min_day=lambda x: pd.to_datetime(x.Min_day),
    )
    return df


def run_rec_freq(
    spark, ho_win=14, model_win=90, sample_ids=[1], ho_start="2018-08-01"
):
    df_spk, q = run_rec_freq_spk(
        spark,
        ho_win=ho_win,
        model_win=model_win,
        sample_ids=sample_ids,
        ho_start=ho_start,
    )
    df = rec_freq_spk2pandas(df_spk)
    return df, df_spk, q



import datetime as dt
import pandas as pd
from pandas.compat import lrange
from pandas import DataFrame

from bgbb.sql.sql_utils import run_rec_freq_spk, rec_freq_spk2pandas, to_s3_fmt
from pytest import fixture


MODEL_START_DATE = dt.date(2018, 6, 1)
MAX_OPPORTUNITIES = 9
DURATION = 10


def int2date(i):
    return MODEL_START_DATE + dt.timedelta(days=i - 1)


rfn_cols = ["Recency", "Frequency", "N"]


@fixture
def clients_daily_df():
    """
    Testing model window of 10 day.
    Day 0: day before model start date
    Day 1: first day of model range
    Day 10: last day of model range
    Day 11: first day *after* model range ('holdout start date')
    """
    #
    client_dates_i = [
        ("Beginning_end", [1, 3, 10]),
        ("Beginning_end_extra", [0, 1, 3, 10, 11]),
        (
            "Before_beginning",
            [0],
        ),  # Not included, since this is before model range
        ("Halfway_start", [5, 6, 11]),
        ("Every_day", lrange(11)),
        ("One_time_returner", [3, 4]),
        ("No_returns", [3]),
    ]

    client_dates = [
        (client_name, [int2date(i) for i in ints])
        for client_name, ints in client_dates_i
    ]

    clients_daily_df = DataFrame(
        [
            (client_name, to_s3_fmt(date))
            for client_name, dates in client_dates
            for date in dates
        ],
        columns=["client_id", "submission_date_s3"],
    )

    clients_daily_df["app_name"] = "Firefox"
    clients_daily_df["channel"] = "release"
    clients_daily_df["sample_id"] = "1"

    return clients_daily_df


@fixture
def rfn_spk(clients_daily_df, spark):
    clients_daily_dfs = spark.createDataFrame(clients_daily_df)

    clients_daily_dfs.createOrReplaceTempView("clients_daily")
    rfn_dfs, _q = run_rec_freq_spk(
        spark, ho_win=1, model_win=10, sample_ids=[1], ho_start="2018-06-11"
    )
    return rfn_dfs


@fixture
def rfn(rfn_spk):
    rfn = (
        rfn_spk.toPandas()
        .set_index("client_id")
        .assign(
            Max_day=lambda x: pd.to_datetime(x.Max_day).dt.date,
            Min_day=lambda x: pd.to_datetime(x.Min_day).dt.date,
        )
    )
    return rfn


def test_user_active_before_range_not_in_rfn(rfn):
    assert "Before_beginning" not in rfn.index
    assert "Beginning_end" in rfn.index
    # User activity outside of range shouldn't make a difference
    # to RFN data
    assert (
        rfn.loc["Beginning_end", rfn_cols]
        == rfn.loc["Beginning_end_extra", rfn_cols]
    ).all()


def test_every_day_user(rfn):
    """A user who shows up every day in a 10-day
    window should have a value of 9 for each value
    of R/F/N.
    """
    assert (rfn.loc["Every_day", rfn_cols] == MAX_OPPORTUNITIES).all()
    assert rfn.loc["Halfway_start"].N < MAX_OPPORTUNITIES


def test_date_range(rfn, clients_daily_df):
    "Dates in rfn don't extend to max/min dates in clients_daily"
    clients_daily_dates = pd.to_datetime(
        clients_daily_df.submission_date_s3
    ).dt.date

    assert clients_daily_dates.min() == int2date(0)
    assert rfn.Min_day.min() == int2date(1)

    assert clients_daily_dates.max() == int2date(11)
    assert rfn.Max_day.max() == int2date(10)


def test_rfn_invariants(rfn, duration=None):
    """
    Assert invariants based on definitions of R/F/N.
    """
    assert MAX_OPPORTUNITIES == DURATION - 1
    assert rfn.eval("Recency <= N").all()
    assert rfn.eval("Recency >= Frequency").all()
    assert (
        rfn.Frequency.eq(0) == rfn.Recency.eq(0)
    ).all(), "For those who never return, Frequency == Recency == 0"

    max_dur = (
        rfn.eval("Max_day - Min_day").astype("timedelta64[D]").astype(int).max()
    )
    assert max_dur == MAX_OPPORTUNITIES


def test_rec_freq_spk2pandas(rfn_spk):
    return rec_freq_spk2pandas(rfn_spk)



import numpy as np
import scipy.stats as st
from lifetimes.datasets import load_donations
from pytest import fixture

from bgbb.bgbb_utils import gen_buy_die, AbgdParams
from bgbb.core import BGBB
from bgbb.wrappers import unload

df = fixture(load_donations)
bg = fixture(lambda: BGBB(params=[1.20, 0.75, 0.66, 2.78]))
pars = fixture(lambda: [1.20, 0.75, 0.66, 2.78])


def test_len(df):
    assert len(df) == 22


def test_cond_prob_alive_nb(df, bg, pars):
    nbres = bg.rfn.cond_prob_alive(df, params=pars, nb=True)
    res = bg.rfn.cond_prob_alive(df, params=pars, nb=False)
    assert np.allclose(nbres, res)


def test_cond_prob_exp_rets(df, bg, pars):
    x, tx, n = unload(df, "frequency recency n")
    bg.data = df
    n14_orig = bg.conditional_expected_number_of_purchases_up_to_time(14)
    n14_nb = bg.cond_exp_rets_till_nb(14, x, tx, n, params=pars)
    n14_api = bg.rfn.cond_exp_rets_till(df, n_days_later=14)
    n14_api_nb = bg.rfn.cond_exp_rets_till(df, n_days_later=14, nb=True)
    assert np.allclose(n14_orig, n14_nb)
    assert np.allclose(n14_orig, n14_api)
    assert np.allclose(n14_orig, n14_api_nb)


def gen_samps_exp(bg, abgd, n_opps, n_users, seed=0):
    # Get median p, th from hyper params
    a, b, g, d = abgd
    actual_p = np.median(st.beta(a, b).rvs(100))
    actual_th = np.median(st.beta(g, d).rvs(100))

    pp, th = np.full(n_users, actual_p), np.full(n_users, actual_th)
    df = gen_buy_die(n_opps, n_users, abgd=abgd, p_th=[pp, th], seed=seed)
    p_est, th_est = bg._expec_p_th(*abgd, df.frequency, df.recency, df.n_opps)
    df = df.assign(P_est=p_est, Th_est=th_est)
    return actual_p, actual_th, p_est, th_est, df


def sim_expec_p_th_diffs(bg, abgd=[70.0, 30.0, 25.0, 75.0], seed=0):
    """
    Hacky test to make sure we can recover
    params p and θ from data generated by them. Hopefully, the median
    estimated params from BGBB._expec_p_th (p_est, th_est) aren't far off
    from the actual params (p_actual, th_actual) used to generate them.
    """
    p_actual, th_actual, p_est, th_est, sim_dat = gen_samps_exp(
        bg, abgd, n_opps=90, n_users=10000, seed=seed
    )
    p_diff = np.median(p_est) - p_actual
    assert abs(p_diff) < 0.03
    th_diff = np.median(th_est) - th_actual
    assert abs(th_diff) < 0.03
    return p_diff, th_diff


def test_expec_p_th(bg):
    sim_expec_p_th_diffs(bg, abgd=[25.0, 75.0, 70.0, 30.0], seed=0)
    sim_expec_p_th_diffs(bg, abgd=[70.0, 30.0, 25.0, 75.0], seed=0)


def test_latent_variable_mean(bg):
    abgd = [25.0, 75.0, 70.0, 30.0]
    p_actual, th_actual, p_est, th_est, df = gen_samps_exp(
        bg, abgd, n_opps=90, n_users=10000, seed=0
    )
    bg2 = BGBB(params=abgd)
    p_est2, th_est2 = bg2.latent_variable_mean(
        df.frequency, df.recency, df.n_opps
    )
    assert np.allclose(p_est, p_est2)
    assert np.allclose(th_est, th_est2)


def beta_exp(a, b):
    """
    Expected value of beta distro
    https://en.wikipedia.org/wiki/Beta_distribution
    """
    return a / (a + b)


def test_gen_buy_die():
    """
    Somewhat circular way of testing the data generating process.
    Checks that params used to generate data with `gen_buy_die`
    can be reasonably recovered with fitter. Also checks that p, θ
    generated by `gen_probs` have mean close to what we'd expect.
    """
    a, b, g, d = abgd = [7, 3, 2, 8]
    n_opps = 90
    n_users = 10000
    df = gen_buy_die(n_opps, n_users, abgd=abgd, p_th=None, seed=0)

    # Check that mean generated p, θ values aren't far from expected
    assert abs((beta_exp(a, b) - df["p"].mean())) < 0.01
    assert abs((beta_exp(g, d) - df["th"].mean())) < 0.01

    rfn_df = (
        df.rename(columns={"n_opps": "n"})
        .groupby(["frequency", "recency", "n"])
        .size()
        .reset_index(drop=0)
        .rename(columns={0: "weights"})
    )

    bg = BGBB(penalizer_coef=0.0).rfn.fit(
        rfn_df.rename(columns={"weights": "n_custs"})
    )
    recovered_params = np.array(list(bg.params_.values()))
    assert (recovered_params.round() == abgd).all()


def test_AbgdParams():
    """
    Test repr, __init__, mod_param, from_dct functionality
    of AbgdParams.
    """
    pars = AbgdParams(7, 3, 2, 8)
    assert repr(pars) == "BGBB Hyperparams <α: 7.0, β: 3.0, γ: 2.0, δ: 8.0>"

    pars2 = pars.mod_param(a=lambda x: x + 1, d=lambda x: x / 2)
    assert pars.d == pars2.d * 2
    assert pars.a == pars2.a - 1

    assert AbgdParams.from_dct(pars._asdict()) == pars



# import pytest
import numpy as np
from scipy.special import betaln, logsumexp
from bgbb.numba_special import nb_lbeta, nb_lbeta_vec2, nb_logsumexp


def test_nb_lbeta(niter=1000):
    for _ in range(niter):
        d, g = np.random.rand(2) * 1000
        a = nb_lbeta(d, g)
        b = betaln(d, g)
        assert np.allclose(a, b), "{} != {} ({}, {})".format(a, b, d, g)


def test_nb_lbeta_vec2():
    assert np.allclose(nb_lbeta_vec2(1, np.arange(7)), betaln(1, np.arange(7)))


def test_nb_logsumexp(n_iter=100):
    np.random.seed(0)
    for _ in range(n_iter):
        a = 10 ** (np.random.rand(100) * 10)
        assert np.allclose(nb_logsumexp(a), logsumexp(a))
